/** 
 * \file deltri.cpp
 *  
 * \brief  Implementation of  some methods  of the  class Subdivision,
 * which represents  a Delaunay triangulation and its  dual (i.e., the
 * Voronoi diagram)  of a point  set in \f$E^2\f$ using  the quad-edge
 * data structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

// -------------------------------------------------------------------
//
// For more  details regarding the  quad-edge data structure,  see the
// paper
//
// "Primitives for  the manipulation  of general subdivisions  and the
// computation  of Voronoi  Diagrams",  by Leonidas  Guibas and  Jorge
// Stolfi,  published  in  the  ACM Transactions  on  Graphics,  4(2),
// p. 74-123, 1985.
//
// Jorge Stolfi  made available  his own code  for the  quad-edge data
// structure:
//
// http://www.ic.unicamp.br/~stolfi/EXPORT/software/c/Index.html
//
// -------------------------------------------------------------------

#include "deltri.hpp"      // DelTri
#include "predicate.hpp"   // Predicate

#include <iostream>        // std::cerr, std::endl
#include <cassert>         // assert
#include <cmath>           // fabs
#include <queue>           // queue


/** 
* \defgroup QuadEdgeNameSpace Namespace quadedge.
* @{
*/

/**
* \namespace quadedge
*
* \brief   The  namespace   quadedge  contains   the   definition  and
* implementation of all classes of the quad-edge data structure, which
* can be used to  represent arbitrary planar subdivisions in Euclidian
* space.
*
*/

namespace quadedge {


  /**
   * \fn DelTri::DelTri( unsigned n , double pts[] )
   *
   * \brief Creates  a instance of this class  representing a Delaunay
   * triangulation of a  set of points in \f$E^2\f$  and its dual (the
   * Voronoi diagram).
   *
   * \param n The number of points.
   * \param pts The point coordinates.
   *
   */
  DelTri::DelTri( unsigned n , double pts[] )
  {
    /**
     * Make sure the point set has at least three points.
     */
    assert( n >= 3 ) ;
    
    /**
     * Find the largest absolute coordinate of the given points.
     */
    double max = 0 ;
    
    for ( unsigned i = 0 ; i < n ; i++ ) {
      double x = fabs( pts[ 2 * i     ] ) ;
      double y = fabs( pts[ 2 * i + 1 ] ) ;
      
      if ( x > max ) {
	max = x ;
      }
      
      if ( y > max ) {
	max = y ;
      }
    }
    
    /**
     * Creates the three vertices of the enclosing triangle.
     */
    max *= 3 ;
    
    Point pa(  max ,    0 , 0 ) ;
    Point pb(    0 ,  max , 0 ) ;
    Point pc( -max , -max , 0 ) ;

    /**
     * Keep the  value of  the largest coordinate  of a vertex  of the
     * enclosing  triangle. This value  will be  used in  the incircle
     * test.
     */
    set_max( max ) ;
    
    /**
     * Creates the enclosing triangle.
     */
    create_enclosing_triangle( pa , pb , pc ) ;
    
    /**
     * Insert the given points into the triangulation.
     */
    for ( unsigned i = 0 ; i < n ; i++ ) {
      insert( Point( pts[ 2 * i ] , pts[ 2 * i + 1 ] , 0 ) ) ;  
    }

    //
    // For debugging purposes only:
    //

    remove_enclosing_triangle() ;

    return ;
  }


  /**
   * \fn virtual DelTri::~DelTri()
   *
   * \brief Releases the memory held by an instance of this class.
   */
  DelTri::~DelTri()
  {
    /**
     * Create a queue for storing edges.
     */
    std::queue< Edge* > q ;
    
    /**
     * Initialize it with the starting edge.
     */
    Edge* efirst =  get_starting_edge() ;

    q.push( efirst ) ;
    
    /**
     * Mark the quad-edge of the starting edge as visited.
     */
    efirst->get_owner()->set_mark( true ) ;

    /**
     * Visit all quad-edges.
     */
    while ( !q.empty() ) {
      /**
       * Get the first edge of the queue.
       */
      Edge* e = q.front() ;
      q.pop() ;
      
      /**
       * Insert adjacent edges into the queue.
       */
      Edge* eaux ;
      
      eaux = e->o_next() ;
      if ( !eaux->get_owner()->is_marked() ) {
	q.push( eaux ) ;
	eaux->get_owner()->set_mark( true ) ;
      }

      eaux = e->o_prev() ;
      if ( !eaux->get_owner()->is_marked() ) {
	q.push( eaux ) ;
	eaux->get_owner()->set_mark( true ) ;
      }

      eaux = e->d_next() ;
      if ( !eaux->get_owner()->is_marked() ) {
	q.push( eaux ) ;
	eaux->get_owner()->set_mark( true ) ;
      }

      eaux = e->d_prev() ;
      if ( !eaux->get_owner()->is_marked() ) {
	q.push( eaux ) ;
	eaux->get_owner()->set_mark( true ) ;
      }

      /**
       * Remove the first quad-edge of the queue. 
       */
      delete_edge( e ) ;

    }

    return ;
  }

  
  /**
   * \fn void DelTri::create_enclosing_triangle( const Point& p1 , const Point& p2 , const Point& p3 )
   *
   * \brief  Creates an enclosing  triangle for  the vertices  of this
   * Delaunay triangulation.  This triangle will be  removed after the
   * triangulation is built.
   *
   * \param p1 First vertex of the enclosing triangle in a CCW traverse.
   * \param p2 Second vertex of the enclosing triangle in a CCW traverse.
   * \param p3 Third  vertex of the enclosing triangle in a CCW traverse.
   *
   */
  void
  DelTri::create_enclosing_triangle(
				    const Point& p1 ,
				    const Point& p2 ,
				    const Point& p3 
				   )
  {
    /**
     * Creates three vertices from the three given ones.
     */
    Point* pa = new Point( p1 ) ;
    Point* pb = new Point( p2 ) ;
    Point* pc = new Point( p3 ) ;

    /**
     * Creates  the first edge,  which connects  the first  and second
     * vertices.
     */
    QuadEdge* qea = new QuadEdge() ;
    Edge* ea = qea->get_edge( 0 ) ;
    ea->endpoints( pa , pb ) ;

    /**
     * Creates the  second edge, which  connects the second  and third
     * vertices.
     */
    QuadEdge* qeb = new QuadEdge() ;
    Edge* eb = qeb->get_edge( 0 ) ;
    eb->endpoints( pb , pc ) ;
    
    /**
     * Attach the first two edge at their common vertex.
     */
    ea->sym()->splice( eb ) ;

    /**
     * Creates the  third edge, which  connecting the first  and third
     * vertices,   giving  rise   to   the  first   triangle  of   the
     * triangulation.
     */
    QuadEdge* qec = new QuadEdge() ;
    Edge* ec = qec->get_edge( 0 ) ;
    ec->endpoints( pc , pa ) ;
    
    /**
     * Attach the third edge to the second one at their common vertex.
     */
    eb->sym()->splice( ec ) ;
    
    /**
     * Attach the third edge to the first one at their common vertex.
     */
    ec->sym()->splice( ea ) ;      
    
    /**
     * Make the  first edge of the  triangle the starting  edge of the
     * triangulation.
     */
    set_starting_edge( ea ) ; 

    /**
     * Set  up  the  edge  points  of the  first,  second,  and  third
     * vertices.
     */
    pa->set_edge( ea ) ;
    pb->set_edge( eb ) ;
    pc->set_edge( ec ) ;

    /**
     * Add  the  new edges  to  the list  of  edges  of this  Delaunay
     * triangulation.
     */
    add_edge( ea ) ;  
    add_edge( eb ) ; 
    add_edge( ec ) ; 

    add_edge( ea->sym() ) ;  
    add_edge( eb->sym() ) ; 
    add_edge( ec->sym() ) ; 

    /**
     * Add the  new points  to the list  of vertices of  this Delaunay
     * triangulation.
     */
    add_point( pa ) ;
    add_point( pb ) ;
    add_point( pc ) ;
    
    return ;
  }


  /**
   * \fn void DelTri::remove_enclosing_triangle()
   *
   * \brief Remove the enclosing triangle and all triangles adjacent
   * to each one  of its vertices (as these  triangles do not belong
   * to  the  Delaunay  triangulation  of  the given  input  set  of
   * points).
   */
  void
  DelTri::remove_enclosing_triangle()
  {
    /**
     * Find an edge of the enclosing triangle.
     */
    Edge* e_next = find_enclosing_triangle_edge() ;

    for ( unsigned i = 0 ; i < 3 ; i++ ) {
      /**
       * Set a  pointer to a boundary  edge incident to  the vertex of
       * the enclosing triangle to  be removed along with its adjacent
       * triangles.
       */
      Edge* e_iter = e_next ;

      /**
       * Get  the another  boundary edge  of the  triangulation before
       * removing a vertex of  the enclosing triangle and its adjacent
       * triangles.
       */
      e_next = e_next->sym()->o_next() ;

      /**
       * Remove a  vertex of the enclosing triangles  and its adjacent
       * triangles.
       */
      remove_boundary_vertex( e_iter ) ;
    }

    return ;
  }


  /**
   * \fn Edge* DelTri::locate( const Point& p ) const
   *
   * \brief Search for an edge  that contains a given point or belongs
   * to a triangle  that contains the given point.   The search starts
   * with the starting edge of this Delaunay triangulation.
   *
   * \param p A given point.
   *
   * \return A pointer to an edge of the triangulation.
   *
   */
  Edge*
  DelTri::locate( const Point& p ) const
  {
    /**
     * Starts the search from the starting edge of the triangulation.
     */
    Edge* e = get_starting_edge() ;
    
    /**
     * Loop  over  the triangles  of  this triangulation  (indirectly)
     * until a triangle containing the given point is found. Note that
     * all points are supposed to be inside the enclosing triangle, so
     * this  loop  should  always  terminate  (unless  some  numerical
     * problem occurs).
     */
    bool end = false ;

    while ( !end ) {
      /**
       * If the  given point coincides with the  origin or destination
       * vertex of edge "e"
       */
      if ( ( p == *( e->get_orig() ) ) ||( p == *( e->get_dest() ) ) ) {
	return e ;
      }
      else if ( right_of( p , e ) ) {
	/**
	 * If the point is on the right face of edge "e", consider the
	 * symmetric of  "e", and repeat  the test for  this symmetric
	 * edge.
	 */
	e = e->sym() ;
      }
      else if ( !right_of( p , e->o_next() ) ) {
	/**
	 * The point is not on the  right face of "e".  But, it is not
	 * in the left face of "e" either. So, consider let us try the
	 * next face  in a CCW traverse  of the faces  incident to the
	 * origin vertex of "e",  starting with the face following the
	 * left face of "e".
	 */
	e = e->o_next() ;
      }
      else if ( !right_of( p , e->d_prev() ) ) {
	/**
	 * The point is on the "cone" defined by "e" and its next edge
	 * with the  same origin vertex. However, the  point is beyond
	 * the third edge of the  triangle containing "e" and the next
	 * edge. In this case, we consider the right face of the third
	 * edge.
	 */
	e = e->d_prev() ;
      }
      else {
	end = true ;
      }
    }

    return e ;
  }


  /**
   * \fn void DelTri::insert( const Point& p )
   *
   * \brief Insert a new vertex into the current triangulation.
   *
   * \param p A given vertex.
   *
   */
  void
  DelTri::insert( const Point& p )
  {
    Edge* e = locate( p ) ;

    /**
     * If the point is already in the triangulation, it is discarded.
     */
    if ( ( p == *( e->get_orig() ) ) ||( p == *( e->get_dest() ) ) ) {
      return ;
    }
    else if ( on_edge( p , e ) ) {
      /**
       * The point  is on an edge,  and we remove the  edge. Note that
       * the removed edge might be the starting edge. So, the starting
       * edge pointer is updated later.
       */
      e = e->o_prev() ;
      delete_edge( e->o_next() ) ;
    }

    /**
     * Connect  the  point  to   the  vertices  of  the  triangle  (or
     * quadrilateral) that  contains it.  To do that,  we create three
     * (or four) new edges.
     */
    QuadEdge* qeb = new QuadEdge() ;
    Edge* eb = qeb->get_edge( 0 ) ;

    /**
     * Add the  new edge (and its  symmetric) to the list  of edges of
     * this Delaunay triangulation.
     */
    add_edge( eb ) ;
    add_edge( eb->sym() ) ;

    /**
     * Create the new point.
     */
    Point* newpt = new Point( p ) ;

    /**
     * Add  the new point  to the  list of  vertices of  this Delaunay
     * triangulation.
     */
    add_point( newpt ) ;

    /**
     * Set up the coordinates of the first edge vertices.
     */
    eb->endpoints( e->get_orig() , newpt ) ;

    /**
     * Set up the edge pointer of the new point.
     */
    newpt->set_edge( eb->sym() ) ;

    /**
     * Attach the first edge to the triangulation.
     */
    eb->splice( e ) ;

    /**
     * Update the starting edge value.
     */
    set_starting_edge( eb ) ;

    /**
     * Create the remaining edges of the triangle.
     */
    do {
      /**
       * A triangle is created by connecting the destination vertex of
       * edge "e" to the origin vertex of the symmetric of edge "eb".
       */
      eb = connect( e , eb->sym() ) ;

      /**
       * Update edge "e".
       */
      e = eb->o_prev() ;
    }
    while ( e->l_next() != get_starting_edge() ) ;

    /**
     * Examine suspect edges  to ensure that all edges  in the current
     * triangulation are  edges of  the Delaunay triangulation  of the
     * point set.
     */
    do {
      /**
       * Get an edge in the right face of edge "e".
       */
      Edge* t = e->o_prev() ;

      /**
       * If the right face is not  an unbounded face, then the face is
       * a triangle incident to  edge "e", which forms a quadrilateral
       * with the left face of "e".  We apply the incircle test to the
       * face.
       */
      if ( 
	  right_of( 
		   *( t->get_dest() ) , 
		   e 
		  ) 
	  &&
	  incircle( 
		   *( e->get_orig() ) ,
		   *( t->get_dest() ) ,
		   *( e->get_dest() ) ,
		   p 
		  ) 
	 ) 
      {
	/**
	 * Swap edge "e".
	 */
	swap( e ) ;
	
	/**
	 * Update  the variable "e",  so it  gets the  another suspect
	 * edge.
	 */
	e = e->o_prev() ;
      }
      else if ( e->o_next() == get_starting_edge() ) {
	/**
	 * If the tested edge passes the incircle test and there is no
	 * other edge to be tested, then all edges are in the Delaunay
	 * triangulation.
	 */
	return ;
      }
      else {
	/**
	 * Get the  next suspect edge. This  edge is obtained  in a CW
	 * traverse of the boundary of the star polygon defined by the
	 * union of all triangles incident to the newly inserted point
	 * "p".
	 */
	e = e->o_next()->l_prev() ;
      }
    }
    while ( true ) ;

    return ;
  }


  /**
   * \fn Edge* DelTri::connect( Edge* ea , Edge* eb ) 
   *
   * \brief  Add a  new edge  to the  triangulation by  connecting the
   * destination  vertex   of  an  edge   to  the  origin   vertex  of
   * another. All three edges will have the same left face after being
   * connected.
   *
   * \param ea Pointer to an edge.
   * \param eb Pointer to an edge.
   *
   * \return Pointer the newly created edge.
   *
   */
  Edge* 
  DelTri::connect( 
		  Edge* ea ,
		  Edge* eb 
		 )
  {
    /**
     * Create a new quad-edge.
     */
    QuadEdge* qe = new QuadEdge() ;
    
    /**
     * Get the first primal edge of the new quad-edge.
     */
    Edge* newe = qe->get_edge( 0 ) ;

    /**
     * Add the new edge to the list of edges of this triangulation.
     */
    add_edge( newe ) ;
    add_edge( newe->sym() ) ;
    
    /**
     * Connect the new edge to edge "ea".
     */
    newe->splice( ea->l_next() ) ;
    
    /**
     * Connect the new edge to edge "eb".
     */
    newe->sym()->splice( eb ) ;
    
    /**
     * Assign vertices to the new edge.
     */
    newe->endpoints( ea->get_dest() , eb->get_orig() ) ;
    
    return newe ;
  }


  /**
   * \fn void DelTri::swap( Edge* e ) 
   *
   * \brief   Turn   a   given   edge  counterclockwise   inside   the
   * quadrilateral given by the two triangles incident to the edge.
   *
   * \param e Pointer to an edge.
   *
   */
  void
  DelTri::swap( Edge* e ) 
  {
    /**
     * Get an edge that shares  the origin vertex with the given edge,
     * and  another one  that shares  the destination  vertex  (in the
     * quadrilateral  defined by  the  two triangles  incident to  the
     * given edge).
     */
    Edge* ea = e->o_prev() ;
    Edge* eb = e->sym()->o_prev() ;

    /**
     * Adjust the  edge pointers  of the origin  vertices of  "ea" and
     * "eb" to point  to those edges (just in case  they point to edge
     * "e").
     */
    ea->get_orig()->set_edge( ea ) ;
    eb->get_orig()->set_edge( eb ) ;
    
    /**
     * Dettach the given  edge from the edges sharing  the same origin
     * vertex.
     */
    e->splice( ea ) ;

    /**
     * Dettach  the  given  edge  from  the  edges  sharing  the  same
     * destination vertex.
     */
    e->sym()->splice( eb ) ;

    /**
     * Attach  the   new  edge   to  the  other   two  edges   of  the
     * quadrilateral.
     */
    e->splice( ea->l_next() ) ;
    e->sym()->splice( eb->l_next() ) ;

    /**
     * Update the origin and destination vertices of the new edge.
     */
    e->endpoints( ea->get_dest() , eb->get_dest() ) ;

    return ;
  }


  /**
   * \fn void DelTri::delete_edge( Edge* e ) 
   *
   * \brief Removes a given edge from the Delaunay triangulation.
   *
   * \param e Pointer to an edge.
   *
   */
  void 
  DelTri::delete_edge( Edge* e )
  {
    /**
     * Get the origin and destination vertices of edge "e".
     */
    Point* po = e->get_orig() ;
    Point* pd = e->get_dest() ;

    /**
     * If the vertices of edge "e"  do not belong to any other edge of
     * the   triangulation,  the   memory   held  by   them  must   be
     * released. Otherwise,  their edge  pointers must point  to other
     * incident edges.
     */
    if ( e != e->o_prev() ) {
      po->set_edge( e->o_prev() ) ;
      po = 0 ;
    }

    if ( e != e->d_prev() ) {
      pd->set_edge( e->sym()->o_prev() ) ;
      pd = 0 ;
    }


    /**
     * Dettach the edge from the stars of its vertices.
     */
    e->splice( e->o_prev() ) ;
    e->sym()->splice( e->sym()->o_prev() ) ;

    /**
     * Remove the edge and its symmetric from the list of edges of the
     * triangulation they belong to.
     */
    _ledges.erase( e->get_position() ) ;
    _ledges.erase( e->sym()->get_position() ) ;

    delete e->get_owner() ;

    /**
     * If  needed, release  the memory  held  by the  vertices of  the
     * deleted edge.
     */
    if ( po != 0 ) {
      _lpoints.erase( po->get_position() ) ;
      delete po ;
    }

    if ( pd != 0 ) { 
      _lpoints.erase( pd->get_position() ) ;
      delete pd ;
    }
  }


  /**
   * \fn bool DelTri::right_of( const Point& p , Edge* e ) const
   *
   * \brief Returns the  Boolean value true if and only  if a point is
   * on the right face of a given edge.
   *
   * \param p A given point.
   * \param e A given edge.
   *
   * \return The Boolean value true if the given point is on the right
   * face of the given edge, and false otherwise.
   *
   */
  bool 
  DelTri::right_of( const Point& p , Edge* e ) const
  {
    Point po( *( e->get_orig() ) ) ;
    Point pd( *( e->get_dest() ) ) ;

    bool res = Predicate::left_on( 
			       po.x() , po.y() ,
			       pd.x() , pd.y() ,
			        p.x() ,  p.y() 
			      ) ;
    
    return !res ;
  }


  /**
   * \fn bool DelTri::on_edge( const Point& p , Edge* e ) const 
   *
   * \brief  Returns the Boolean  value true  if and  only if  a point
   * belongs to an edge.
   *
   * \param p A given point.
   * \param e A given edge.
   *
   * \return The Boolean value true  if the given point belongs to the
   * given edge, and false otherwise.
   *
   */
  bool 
  DelTri::on_edge( const Point& p , Edge* e ) const
  {
    Point po = *( e->get_orig() ) ;
    Point pd = *( e->get_dest() ) ;
    
    Predicate::ORIENTATION orient = 
      Predicate::classify( 
			  po.x() ,
			  po.y() ,
			  pd.x() ,
			  pd.y() ,
			  p.x() ,
			  p.y()
			 ) ;
    
    return ( orient == Predicate::ORIGIN ) || 
      ( orient == Predicate::DESTINATION ) ||
      ( orient == Predicate::BETWEEN ) ;
  }


  /**
   * \fn bool DelTri::incircle( const Point& a , const Point& b , const Point& c , const Point& d) const 
   *
   * \brief  Returns the Boolean  value true  if and  only if  a given
   * point d is  in the interior of a circle  defined by three points:
   * a, b, and c.
   *
   * \param a First point defining a circle.
   * \param b Second point defining a circle.
   * \param c Third point defining a circle.
   * \param d A point to be tested against the incircle test.
   *
   * \return  The Boolean  value  true if  the  test point  is in  the
   * interior of  the circle  defined by the  other three  points, and
   * false otherwise.
   *
   */
  bool
  DelTri::incircle(
		   const Point& a , 
		   const Point& b , 
		   const Point& c , 
		   const Point& d
		  ) 
    const
  {
    /**
     * To  make sure that  the enclosing  triangle does  not interfere
     * with the Delaunay triangulation,  we use the strategy suggested
     * in the  book "Computational Geometry", by de  Berg, Cheong, van
     * Kreveld, and Overmars, 3rd edition, 2008, page 204.
     */
    unsigned rank_a = 0 ;
    if ( fabs( a.x() ) == get_max() ) rank_a = 1 ;
    if ( fabs( a.y() ) == get_max() ) rank_a = rank_a + 2 ;
    
    unsigned rank_b = 0 ;
    if ( fabs( b.x() ) == get_max() ) rank_b = 1 ;
    if ( fabs( b.y() ) == get_max() ) rank_b = rank_b + 2 ;
    
    unsigned rank_c = 0 ;
    if ( fabs( c.x() ) == get_max() ) rank_c = 1 ;
    if ( fabs( c.y() ) == get_max() ) rank_c = rank_c + 2 ;
    
    if ( ( rank_a == 0 ) && ( rank_b == 0 ) && ( rank_c == 0 ) ) {
      return Predicate::incircle( 
				 a.x() , a.y() , b.x() , b.y() , 
				 c.x() , c.y() , d.x() , d.y() 
				) ;
    }
    else if ( ( rank_b > rank_c ) && ( rank_b > rank_a ) ) { 
      return false ;
    }

    /**
     * If we reach this, then the edge "ac" is incident to a vertex of
     * the enclosing triangle and we  want to swap this edge. However,
     * we can only do that if quadrilateral "abcd" is strictly convex. 
     */
    return Predicate::left( b.x() , b.y() , c.x() , c.y() , d.x() , d.y() ) &&
      !Predicate::left_on( b.x() , b.y() , a.x() , a.y() , d.x() , d.y() ) ;
  }


  /**
   * \fn Edge* DelTri::find_enclosing_triangle_edge() const
   *
   * \brief Find one edge of the enclosing triangle.
   *
   * \return A pointer to an edge of the enclosing triangle.
   *
   */
  Edge*
  DelTri::find_enclosing_triangle_edge() const
  {
    /**
     * Create a point with coordinates ( max , 0 ).
     */
    Point po( get_max() , 0 , 0 ) ;

    /**
     * Find an edge incident on this point.
     */
    Edge* e = locate( po ) ;

    /**
     * If the edge edge is not an edge of the enclosing triangle, then
     * find such an edge by traversing the star of the vertex at point
     * po.
     */

    /**
     * Make sure po  is the origin vertex  of e. If not, let  e be its
     * mate.
     */
    if ( e->get_orig()->x() != get_max() ) {
      e = e->sym() ;

      /**
       * If  po is not  the origin  vertex of  e, then  something went
       * wrong with  the point location  method, as it is  expected to
       * return an edge incident to a  vertex at po.  So, we can abort
       * execution.
       */
      assert( e->get_orig()->x() == get_max() ) ;
    }

    /**
     * Traverse the  star of  the vertex at  po until  the destination
     * vertex of the current edge e  is the vertex of coordinates (0 ,
     * max).
     */
    while ( e->get_dest()->y() != get_max() ) {
      e = e->o_next() ;
    }

    return e ;
  }


  /**
   * \fn void DelTri::remove_boundary_vertex( Edge* e )
   *
   * \brief Remove all triangles incident to a vertex of the enclosing
   * triangle, and  then remove the  vertex from the  current Delaunay
   * triangulation.
   *
   * \param e A boundary edge incident to the vertex to be removed.
   */
  void
  DelTri::remove_boundary_vertex( Edge* e )
  {
    /**
     * Get the other boundary edge incident to the boundary vertex.
     */
    Edge* e_iter = e ;
    Edge* e_last = e->o_prev() ;

    /**
     * Visit and delete all edges incident to the boundary vertex.
     */
    bool done = false ;
    while ( !done ) {
      /**
       * Get the next edge in a CCW traverse around the vertex.
       */
      Edge* e_next = e_iter->o_next() ;

      /**
       * Make sure the edge to be removed is not the starting edge.
       */
      if ( 
	  ( e_iter        == get_starting_edge() ) || 
	  ( e_iter->sym() == get_starting_edge() ) 
	 )
      {
	set_starting_edge( e_iter->d_prev() ) ;
      }

      /**
       * If the  next edge in the  traversal is a  boundary edge, then
       * there  is only one  remaining triangle,  and hence  two edges
       * must be removed from  the triangulation.  Otherwise, only one
       * is.
       */
      if ( e_next == e_last ) {
	//
	// For debugging purposes only:
	//
	std::cerr << std::endl
		  << "removing edges: po = ( "
		  << e_iter->get_orig()->x() 
		  << " , "
		  << e_iter->get_orig()->y() 
		  << " ) and pd: ( "
		  << e_iter->get_dest()->x() 
		  << " , "
		  << e_iter->get_dest()->y() 
		  << " ) and po = ( "
		  << e_last->get_orig()->x() 
		  << " , "
		  << e_last->get_orig()->y() 
		  << " ) and pd: ( "
		  << e_last->get_dest()->x() 
		  << " , "
		  << e_last->get_dest()->y() 
		  << " )."
		  << std::endl ;

	remove_boundary_triangle_and_vertex( e_iter , e_last ) ;

	done = true ;
      }
      else {
	//
	// For debugging purposes only:
	//
	std::cerr << std::endl
		  << "removing edge: po = ( "
		  << e_iter->get_orig()->x() 
		  << " , "
		  << e_iter->get_orig()->y() 
		  << " ) and pd: ( "
		  << e_iter->get_dest()->x() 
		  << " , "
		  << e_iter->get_dest()->y() 
		  << " ) ."
		  << std::endl ;

	remove_boundary_triangle( e_iter ) ;

	e_iter = e_next ;
      }
    }

    return ;
  }


  /**
   * \fn void DelTri::remove_boundary_triangle( Edge* e )
   *
   * \brief Remove a triangle incident to exactly one boundary edge.
   *
   * \param e A boundary edge incident to the triangle.
   */
  void
  DelTri::remove_boundary_triangle( Edge* e )
  {
    delete_edge( e ) ;

    return ;
  }


  /**
   * \fn void DelTri::remove_boundary_triangle_and_vertex( Edge* e1 , Edge* e2 )
   *
   * \brief Remove a triangle incident to exactly two boundary edges.
   *
   * \param e1 Pointer to a boundary edge of the triangle.
   * \param e2 Pointer to the other boundary edge of the triangle.
   */
  void
  DelTri::remove_boundary_triangle_and_vertex( Edge* e1 , Edge* e2 )
  {
    delete_edge( e1 ) ;
    delete_edge( e2 ) ;

    return ;
  }


}

/** @} */ //end of group class.
