/** 
 * \file deltri.hpp
 *  
 * \brief  Definition and  implementation of  the class  DelTri, which
 * represents a Delaunay triangulation and its dual (i.e., the Voronoi
 * diagram)  of a  point set  in  \f$E^2\f$ using  the quad-edge  data
 * structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

// -------------------------------------------------------------------
//
// For more  details regarding the  quad-edge data structure,  see the
// paper
//
// "Primitives for  the manipulation  of general subdivisions  and the
// computation  of Voronoi  Diagrams",  by Leonidas  Guibas and  Jorge
// Stolfi,  published  in  the  ACM Transactions  on  Graphics,  4(2),
// p. 74-123, 1985.
//
// Jorge Stolfi  made available  his own code  for the  quad-edge data
// structure:
//
// http://www.ic.unicamp.br/~stolfi/EXPORT/software/c/Index.html
//
// -------------------------------------------------------------------

#ifndef DELTRI_HPP
#define DELTRI_HPP

#include "visitor.hpp"     // Visitor
#include "point.hpp"       // Point
#include "edge.hpp"        // Edge
#include "quadedge.hpp"    // QuadEdge

#include <vector>          // std::vector
#include <list>            // std::list


/** 
* \defgroup QuadEdgeNameSpace Namespace quadedge.
* @{
*/

/**
* \namespace quadedge
*
* \brief   The  namespace   quadedge  contains   the   definition  and
* implementation of all classes of the quad-edge data structure, which
* can be used to  represent arbitrary planar subdivisions in Euclidian
* space.
*
*/

namespace quadedge {

  /**
   * \class DelTri
   *
   * \brief This class represents  a Delaunay triangulation of a point
   * set  in  \f$E^2\f$  (and  its  dual)  using  the  quad-edge  data
   * structure.
   */
  class DelTri {
  public:

    // ---------------------------------------------------------------
    //
    // Type definitions
    //
    // ---------------------------------------------------------------

    /**
     * \typedef PointIterator
     *
     * \brief Defines  an iterator for  the vertices of  this Delaunay
     * triangulation.
     */
    typedef std::list< Point* >::const_iterator PointIterator ;


    /**
     * \typedef EdgeIterator
     *
     * \brief  Defines an  iterator  for the  edges  of this  Delaunay
     * triangulation.
     */
    typedef std::list< Edge* >::const_iterator EdgeIterator ;


    /**
     * \typedef FaceIterator
     *
     * \brief Defines  an iterator for the triangles  of this Delaunay
     * triangulation.
     */
    // typedef typename std::list< Face* >::const_iterator FaceIterator ;


    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn DelTri( unsigned n , double pts[] )
     *
     * \brief Creates a instance of this class representing a Delaunay
     * triangulation of a set of points in \f$E^2\f$ and its dual (the
     * Voronoi diagram).
     *
     * \param n The number of points.
     * \param pts The point coordinates.
     *
     */
    DelTri( unsigned n , double pts[] ) ;


    /**
     * \fn Edge* get_starting_edge() const 
     *
     * \brief Returns the starting edge of this triangulation.
     *
     * \return The starting edge of this triangulation.
     *
     */
    Edge* get_starting_edge() const { return _sedge ; }


    /**
     * \fn double get_max() const 
     *
     * \brief Returns the absolute  value of the largest coordinate of
     * a vertex of the enclosing triangle.
     *
     * \return  The absolute  value  of the  largest  coordinate of  a
     * vertex of the enclosing triangle.
     *
     */
    double get_max() const { return _max ; }


    /**
     * \fn unsigned get_number_of_vertices() const
     *
     * \brief  Returns  the  number   of  vertices  of  this  Delaunay
     * triangulation.
     *
     * \return The number of vertices of this Delaunay triangulation.
     *
     */
    unsigned get_number_of_vertices() const { return _lpoints.size() ; }


    /**
     * \fn unsigned get_number_of_edges() const
     *
     * \brief   Returns  the   number  of   edges  of   this  Delaunay
     * triangulation.
     *
     * \return The number of edges of this Delaunay triangulation.
     *
     */
    unsigned get_number_of_edges() const { return _ledges.size() >> 1 ; }


    /**
     * \fn virtual ~DelTri()
     *
     * \brief Releases the memory held by an instance of this class.
     */
    virtual ~DelTri() ;



    /**
     * \fn PointIterator points_begin() const
     *
     * \brief Returns a point iterator pointing to the first vertex of
     * the list of vertices of this triangulation.
     *
     * \return A point  list iterator pointing to the  first vertex of
     * the list.
     */
    PointIterator points_begin() const
    { 
      return _lpoints.begin() ;
    }


    /**
     * \fn EdgeIterator edges_begin() const 
     *
     * \brief Returns an  edge iterator pointing to the  first edge of
     * the list of edges of this triangulation.
     *
     * \return An edge list iterator pointing to the first edge of the
     * list.
     */
    EdgeIterator edges_begin() const
    {
      return _ledges.begin() ;  
    }


    /**
     * \fn PointIterator points_end() const
     *
     * \brief Returns a point iterator  pointing to the last vertex of
     * the list of vertices of this triangulation.
     *
     * \return A point list iterator  pointing past the last vertex of
     * the list.
     */
    inline PointIterator points_end() const
    { 
      return _lpoints.end() ;
    }


    /**
     * \fn inline EdgeIterator edges_end() const 
     *
     * \brief Returns an edge iterator  pointing past the last edge of
     * the list of edges of this subdivision.
     *
     * \return An  edge list iterator  pointing past the last  edge of
     * the list.
     */
    inline EdgeIterator edges_end() const
    {
      return _ledges.end() ;  
    }


  protected : 

    // ---------------------------------------------------------------
    //
    // Protected methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn void set_starting_edge( Edge* e )
     *
     * \brief Assigns an  address to the pointer to  the starting edge
     * of this triangulation.
     *
     * \param e The address of an edge.
     *
     */
    void set_starting_edge( Edge* e ) { 
      _sedge = e ; 
    }


    /**
     * \fn void set_max( double val ) 
     *
     * \brief Set a value to the largest coordinate of a vertex of the
     * enclosing triangle.
     *
     * \param val Maximum absolute valute of the largest coordinate of
     * a vertex of the enclosing triangle.
     *
     */
    void set_max( double val )  { _max = val ; }


    /**
     * \fn void create_enclosing_triangle( const Point& p1 , const Point& p2 , const Point& p3 )
     *
     * \brief Creates  an enclosing triangle for the  vertices of this
     * Delaunay  triangulation.  This triangle  will be  removed after
     * the triangulation is built.
     *
     * \param  p1 First  vertex of  the  enclosing triangle  in a  CCW
     * traverse.
     * \param  p2 Second  vertex of  the enclosing  triangle in  a CCW
     * traverse.
     * \param  p3 Third  vertex of  the  enclosing triangle  in a  CCW
     * traverse.
     *
     */
    void create_enclosing_triangle( const Point& p1 , const Point& p2 ,
      const Point& p3 ) ;


    /**
     * \fn void insert( const Point& p )
     *
     * \brief Insert  a new vertex into the  current triangulation.
     *
     * \param p A given vertex.
     *
     */
    void insert( const Point& p ) ;


    /**
     * \fn void remove_enclosing_triangle()
     *
     * \brief Remove the enclosing triangle and all triangles adjacent
     * to each one  of its vertices (as these  triangles do not belong
     * to  the  Delaunay  triangulation  of  the given  input  set  of
     * points).
     *
     */
    void remove_enclosing_triangle() ;


    /**
     * \fn Edge* locate( const Point& p ) const
     *
     * \brief  Search for  an  edge  that contains  a  given point  or
     * belongs  to a  triangle  that contains  the  given point.   The
     * search  starts   with  the  starting  edge   of  this  Delaunay
     * triangulation.
     *
     * \param p A given point.
     *
     * \return A pointer to an edge of the triangulation.
     *
     */
    Edge* locate( const Point& p ) const ;


    /**
     * \fn Edge* connect( Edge* ea , Edge* eb ) 
     *
     * \brief Add  a new edge  to the triangulation by  connecting the
     * destination  vertex  of  an   edge  to  the  origin  vertex  of
     * another. All  three edges  will have the  same left  face after
     * being connected.
     *
     * \param ea Pointer to an edge.
     * \param eb Pointer to an edge.
     *
     * \return Pointer the newly created edge.
     *
     */
    Edge* connect( Edge* ea , Edge* eb ) ;


    /**
     * \fn void swap( Edge* e ) 
     *
     * \brief   Turn  a   given  edge   counterclockwise   inside  the
     * quadrilateral given by the two triangles incident to the edge.
     *
     * \param e Pointer to an edge.
     *
     */
    void swap( Edge* e ) ;


    /**
     * \fn void delete_edge( Edge* e ) 
     *
     * \brief Removes a given edge from the Delaunay triangulation.
     *
     * \param e Pointer to an edge.
     *
     */
    void delete_edge( Edge* e ) ;


    /**
     * \fn bool right_of( const Point& p , Edge* e ) const 
     *
     * \brief Returns the Boolean value true if and only if a point is
     * on the right face of a given edge.
     *
     * \param p A given point.
     * \param e A given edge.
     *
     * \return The  Boolean value  true if the  given point is  on the
     * right face of the given edge, and false otherwise.
     *
     */
    bool right_of( const Point& p , Edge* e ) const ;


    /**
     * \fn bool on_edge( const Point& p , Edge* e ) const 
     *
     * \brief Returns  the Boolean value true  if and only  if a point
     * belongs to an edge.
     *
     * \param p A given point.
     * \param e A given edge.
     *
     * \return The  Boolean value true  if the given point  belongs to
     * the given edge, and false otherwise.
     *
     */
    bool on_edge( const Point& p , Edge* e ) const ;
 

    /**
     * \fn bool incircle( const Point& a , const Point& b , const Point& c , const Point& d) const 
     *
     * \brief Returns  the Boolean value true  if and only  if a given
     * point d is in the interior of a circle defined by three points:
     * a, b, and c.
     *
     * \param a First point defining a circle.
     * \param b Second point defining a circle.
     * \param c Third point defining a circle.
     * \param d A point to be tested against the incircle test.
     *
     * \return  The Boolean value  true if  the test  point is  in the
     * interior of the  circle defined by the other  three points, and
     * false otherwise.
     *
     */
    bool incircle( const Point& a , const Point& b , const Point& c ,
      const Point& d) const ;


    /**
     * \fn void add_point( Point* p )
     *
     * \brief  Inserts a  point  into  the list  of  vertices of  this
     * Delaunay triangulation.
     *
     * \param p A pointer to a point.
     *
     */
    void add_point( Point* p ) {
      p->set_position( _lpoints.insert( _lpoints.end() , p ) ) ;
    }


    /**
     * \fn void add_edge( Edge* e )
     *
     * \brief  Inserts an edge  into  the list  of  vertices of  this
     * Delaunay triangulation.
     *
     * \param e A pointer to an edge.
     *
     */
    void add_edge( Edge* e ) {
      e->set_position( _ledges.insert( _ledges.end() , e ) ) ;
    }


    /**
     * \fn Edge* find_enclosing_triangle_edge() const
     *
     * \brief Find one edge of the enclosing triangle.
     *
     * \return A pointer to an edge of the enclosing triangle.
     *
     */
    Edge* find_enclosing_triangle_edge() const ;


    /**
     * \fn void remove_boundary_vertex( Edge* e )
     *
     * \brief  Remove  all  triangles  incident  to a  vertex  of  the
     * enclosing triangle, and then remove the vertex from the current
     * Delaunay triangulation.
     *
     * \param e A boundary edge incident to the vertex to be removed.
     */
    void remove_boundary_vertex( Edge* e ) ;


    /**
     * \fn void remove_boundary_triangle( Edge* e )
     *
     * \brief Remove a triangle incident to exactly one boundary edge.
     *
     * \param e A boundary edge incident to the triangle.
     */
    void remove_boundary_triangle( Edge* e ) ;


    /**
     * \fn void remove_boundary_triangle_and_vertex( Edge* e1 , Edge* e2 )
     *
     * \brief  Remove  a triangle  incident  to  exactly two  boundary
     * edges.
     *
     * \param e1 Pointer to a boundary edge of the triangle.
     * \param e2 Pointer to the other boundary edge of the triangle.
     */
    void remove_boundary_triangle_and_vertex(
					     Edge* e1 ,
					     Edge* e2 
					     ) ;


    // ---------------------------------------------------------------
    //
    // Protected data members
    //
    // ---------------------------------------------------------------

    Edge* _sedge  ;  ///< An edge of this triangulation.
    double _max  ;  ///< Absolute value of the largest (or smallest) coordinate of a vertex of the enclosing triangle.
    std::list< Point* > _lpoints ;  ///< List of vertices of this Delaunay triangulation.
    std::list< Edge* > _ledges ;  ///< List of edges of this Delaunay triangulation.
    // std::list< Face* > _lfaces ;  ///< List of triangles of this Delaunay triangulation.

  } ;


}

/** @} */ //end of group class.

#endif   // DELTRI_HPP
