/** 
 * \file edge.hpp
 *  
 * \brief  Definition  and implementation  of  the  class Edge,  which
 * represents any  of the four edges  of a quad-edge  in the quad-edge
 * data structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

// -------------------------------------------------------------------
//
// For more  details regarding the  quad-edge data structure,  see the
// paper
//
// "Primitives for  the manipulation  of general subdivisions  and the
// computation  of Voronoi  Diagrams",  by Leonidas  Guibas and  Jorge
// Stolfi,  published  in  the  ACM Transactions  on  Graphics,  4(2),
// p. 74-123, 1985.
//
// Jorge Stolfi  made available  his own code  for the  quad-edge data
// structure:
//
// http://www.ic.unicamp.br/~stolfi/EXPORT/software/c/Index.html
//
// -------------------------------------------------------------------

#ifndef EDGE_HPP
#define EDGE_HPP

#include "visitor.hpp" // Visitor
#include "point.hpp"   // Point

#include <cassert>     // assert


/** 
* \defgroup QuadEdgeNameSpace Namespace quadedge.
* @{
*/

/**
* \namespace quadedge
*
* \brief   The  namespace   quadedge  contains   the   definition  and
* implementation of all classes of the quad-edge data structure, which
* can be used to  represent arbitrary planar subdivisions in Euclidian
* space.
*
*/

namespace quadedge {

  /**
   * Forward definition of class QuadEdge.
   */
  class QuadEdge ;


  /**
   * \class Edge
   *
   * \brief This class represents any of the four edges of a quad-edge
   * in the quad-edge data structure.
   */
  class Edge {
  public: 

    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn Edge()
     *
     * \brief Creates an instance of this class.
     *
     */
    Edge() : _owner(0), _id( 0 ) , _next( 0 ) , _orig( 0 ) // , _pos( 0 )
    {}


    /**
     * \fn ~Edge()
     *
     * \brief Releases the memory held by an instance of this class.
     */
    virtual ~Edge()
    {}


    /**
     * \fn QuadEdge* get_owner() const
     *
     * \brief Returns  a pointer to  the quad-edge that  contains this
     * edge.
     *
     * \return A pointer to the quad-edge that contains this edge.
     *
     */
    QuadEdge* get_owner() const
    {
      return _owner ;
    }


    /**
     * \fn unsigned get_id() const
     *
     * \brief Returns the identifier of this edge.
     *
     * \return The identifier of this edge.
     *
     */
    unsigned get_id() const
    {
      return _id ;
    }


    /**
     * \fn Point* get_orig() const
     *
     * \brief Returns the origin vertex of this edge.
     *
     * \return The origin vertex of this edge.
     *
     */
    Point* get_orig() const
    {
      return _orig ;
    }


    /**
     * \fn Point* get_dest() const
     *
     * \brief Returns the destination vertex of this edge.
     *
     * \return The destination vertex of this edge.
     *
     */
    Point* get_dest() const
    {
      return sym()->get_orig() ;
    }


    /**
     * \fn std::list< Edge* >::iterator get_position() const
     *
     * \brief Returns the position of this  edge in a list of edges of
     * the triangulation it belongs to.
     *
     * \return The  position of this  edge in a  list of edges  of the
     * triangulation it belongs to.
     *
     */
    std::list< Edge* >::iterator get_position() const
    {
      return _pos ;
    }


    /**
     * \fn void set_position( const std::list< Edge* >::iterator& it )
     *
     * \brief Assigns a  position for this edge in a  list of edges of
     * the triangulation it belongs to.
     *
     * \param  it An  iterator  position in  a  list of  edges of  the
     * triangulation it belongs to.
     *
     */
    void set_position( const std::list< Edge* >::iterator& it )
    {
      _pos = it ;
    }


    /**
     * \fn Edge* rot() const
     *
     * \brief Returns the dual of this edge.
     *
     * \return The dual of this edge.
     *
     */
    Edge* rot() const
    {
      return ( get_id() < 3 ) ? const_cast< Edge* >( this + 1 ) :
	                        const_cast< Edge* >( this - 3 ) ;
    }


    /**
     * \fn Edge* invrot() const
     *
     * \brief Returns the symmetric of the dual of this edge.
     *
     * \return The symmetric of the dual of this edge.
     *
     */
    Edge* invrot() const
    {
      return ( get_id() > 0 ) ? const_cast< Edge* >( this - 1 ) :
                                const_cast< Edge* >( this + 3 ) ; 
    }


    /**
     * \fn Edge* sym() const
     *
     * \brief Returns the symmetric of this edge.
     *
     * \return The symmetric of this edge.
     *
     */
    Edge* sym() const
    {
      return ( get_id() < 2 ) ? const_cast< Edge* >( this + 2 ) :
	                        const_cast< Edge* >( this - 2 ) ; 
    }


    /**
     * \fn Edge* o_next() const
     *
     * \brief Returns the "next" edge in a CCW traverse of the star of
     * the origin vertex of this edge.
     *
     * \return The  "next" edge in a  CCW traverse of the  star of the
     * origin vertex of this edge.
     *
     */
    Edge* o_next() const
    {
      return _next ; 
    }


    /**
     * \fn Edge* o_prev() const
     *
     * \brief Returns the "next" edge in  a CW traverse of the star of
     * the origin vertex of this edge.
     *
     * \return The  "next" edge in  a CW traverse  of the star  of the
     * origin vertex of this edge.
     *
     */
    Edge* o_prev() const
    {
      return rot()->o_next()->rot() ; 
    }


    /**
     * \fn Edge* d_next() const
     *
     * \brief Returns the "next" edge in a CCW traverse of the star of
     * the destination vertex this edge.
     *
     * \return The  "next" edge in a  CCW traverse of the  star of the
     * destination vertex of this edge.
     *
     */
    Edge* d_next() const
    {
      return sym()->o_next()->sym() ;
    }


    /**
     * \fn Edge* d_prev() const
     *
     * \brief Returns the "next" edge in  a CW traverse of the star of
     * the destination vertex of this edge.
     *
     * \return The  "next" edge in  a CW traverse  of the star  of the
     * destination vertex of this edge.
     *
     */
    Edge* d_prev() const
    {
      return invrot()->o_next()->invrot() ; 
    }


    /**
     * \fn Edge* l_next() const
     *
     * \brief Returns  the "next" edge in  a CCW traverse  of the left
     * face of this edge.
     *
     * \return The "next"  edge in a CCW traverse of  the left face of
     * this edge.
     *
     */
    Edge* l_next() const
    {
      return invrot()->o_next()->rot() ;
    }


    /**
     * \fn Edge* l_prev() const
     *
     * \brief Returns the "previous" edge in a CW traverse of the left
     * face of this edge.
     *
     * \return The "previous"  edge in a CW traverse  of the left face
     * of this edge.
     *
     */
    Edge* l_prev() const
    {
      return o_next()->sym() ; 
    }


    /**
     * \fn Edge* r_next() const
     *
     * \brief Returns the  "next" edge in a CCW  traverse of the right
     * face of this edge.
     *
     * \return The "next" edge in a  CCW traverse of the right face of
     * this edge.
     *
     */
    Edge* r_next() const
    {
      return rot()->o_next()->invrot() ;
    }


    /**
     * \fn Edge* r_prev() const
     *
     * \brief  Returns the  "previous" edge  in a  CW traverse  of the
     * right face of this edge.
     *
     * \return The "previous" edge in  a CW traverse of the right face
     * of this edge.
     *
     */
    Edge* r_prev() const
    {
      return sym()->o_next() ; 
    }


    /**
     * \fn void endpoints( Point* orig , Point* dest )
     *
     * \brief Assigns an origin and a destination vertex to an edge of
     * this quad-edge.
     *
     * \param orig A vertex to be assigned as origin vertex of an edge.
     * \param dest A vertex to be assigned as destination vertex of an
     * edge.
     *
     */
    void endpoints( Point* orig , Point* dest )
    {
     set_orig( orig ) ;
     set_dest( dest ) ;

      return ;
    }


    /**
     * \fn void splice( Edge* e )
     *
     * \brief If  this edge and the  given edge share  the same origin
     * vertex,  then this  method will  break the  star into  two ones
     * (i.e., this edge and the given edge will be separated from each
     * other). Otherwise, the  star of the origin vertex  of this edge
     * and the  star of the  origin vertex of  the given edge  will be
     * combined into one star (i.e., this edge and the given edge will
     * be attached to each other).   In the latter case, the operation
     * will  only  make sense  (in  a  geometric  way) if  the  origin
     * vertices  of  both edges  are  the  "same". But,  topologically
     * speaking, the  operation is  always consistent (see  Guibas and
     * Stolfi paper).
     *
     * \param e A pointer to a given edge.
     *
     */
    void splice( Edge* e )
    {
      Edge* tnextdual = this->o_next()->rot() ;
      Edge* enextdual = e->o_next()->rot() ;

      Edge* temp1 = e->o_next() ;
      Edge* temp2 = this->o_next() ;
      Edge* temp3 = enextdual->o_next() ;
      Edge* temp4 = tnextdual->o_next() ;

      this->set_next( temp1 ) ;
      e->set_next( temp2 ) ;

      tnextdual->set_next( temp3 ) ;
      enextdual->set_next( temp4 ) ;

      return ;
    }


    /**
     * \fn void accept( Visitor* v )
     *
     * \brief Accepts a visit of a generic visitor.
     *
     * \param v A pointer to a generic visitor.
     *
     */
    void accept( Visitor* v )
    {
      v->visit( this ) ;

      return ;
    }


    /**
     * Declare QuadEdge as a friend of this class.
     */
    friend class QuadEdge ;


  protected :

    // ---------------------------------------------------------------
    //
    // Protected methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn void set_owner( QuadEdge* qe )
     *
     * \brief Makes this edge part of a quad-edge.
     *
     * \param qe A pointer to a quad-edge.
     *
     */
    void set_owner( QuadEdge* qe )
    {
      assert( qe != 0 ) ;

      _owner = qe ;
    }


    /**
     * \fn void set_id( unsigned id )
     *
     * \brief Assign an identifier to this edge.
     *
     * \param id An edge identifier (an integer from 0 to 3)
     *
     */
    void set_id( unsigned id )
    {
      assert( id <= 3 ) ;

      _id = id ;
    }


    /**
     * \fn void set_next( Edge* e )
     *
     * \brief Assign a "next" edge to this edge.
     *
     * \param e A pointer to the "next" edge.
     *
     */
    void set_next( Edge* e )
    {
      assert( e != 0 ) ;

      _next = e ;
    }


    /**
     * \fn void set_orig( Point* v )
     *
     * \brief Assign an origin vertex to this edge.
     *
     * \param v A pointer to a vertex.
     *
     */
    void set_orig( Point* v )
    {
      _orig = v ;
    }
    

    /**
     * \fn void set_dest( Point* v )
     *
     * \brief Assign a destination vertex to this edge.
     *
     * \param v A pointer to a vertex.
     *
     */
    void set_dest( Point* v )
    {
      sym()->set_orig( v ) ;
    }


    // ---------------------------------------------------------------
    //
    // Protected data members
    //
    // ---------------------------------------------------------------
    
    QuadEdge* _owner ;  ///< The quad edge that contains this edge.
    
    unsigned _id ;    ///< The id of this edge (an integer from 0 to 3 ).
    
    Edge* _next ;     ///< The address of the "next" edge.
    
    Point* _orig ;    ///< The origin vertex of this edge.
    
    std::list< Edge* >::iterator _pos ;  ///< Position of this edge in a list of edges of the triangulation it belongs to. 
    
  } ;


}

/** @} */ //end of group class.

#endif   // EDGE_HPP
