/** 
 * \file edgevisitor.cpp
 *  
 * \brief Implementation of class EdgeVisitor, which is a visitor that
 * creates a list of all edges of a Delaunay triangulation represented
 * by a quad-edge data structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "edgevisitor.hpp"

#include <cassert>   // assert


/** 
* \defgroup QuadEdgeNameSpace Namespace quadedge.
* @{
*/

/**
* \namespace quadedge
*
* \brief   The  namespace   quadedge  contains   the   definition  and
* implementation of all classes of the quad-edge data structure, which
* can be used to  represent arbitrary planar subdivisions in Euclidian
* space.
*
*/

namespace quadedge {


  /**
   * \fn void EdgeVisitor::visit( DelTri& dt , unsigned& ne , unsigned*& edges , unsigned& nv , double*& verts )
   *
   * \brief A generic visitor method to visit a Delaunay triangulation
   * represented by  the quad-edge data structure.   During the visit,
   * the  method creates  a list  with the  coordinates of  origin and
   * destination vertices of all visited edges.
   *
   * \param dt  A Delaunay triangulation represented  by the quad-edge
   * data structure.
   * \param ne Total number of edges in the triangulation.
   * \param  edges  An array  with  the  indices  of all  edge  vertex
   * coordinates.
   * \param nv Total number of vertices of the triangulation.
   * \param verts  Coordinates of the origin  and destination vertices
   * of all edges.
   *
   */
  void
  EdgeVisitor::visit( 
		     DelTri& dt ,
		     unsigned& ne ,
		     unsigned*& edges ,
		     unsigned& nv ,
		     double*& verts 
		    )
  {
    /**
     * Creates  a temporary  edge list  by visiting  all edges  of the
     * Delaunay triangulation.  This is done by calling the visitor of
     * the  quad-edge containing  the Delaunay  triangulation starting
     * edge.
     */
    _elist.clear() ;
    _vlist.clear() ;

    /**
     * Get the largest  coordinate (in absolute value) of  a vertex of
     * the triangulation enclosing triangle.
     */
    _max = dt.get_max() ;

    /**
     * Visit the first edge of the Delaunay triangulation.
     */
    dt.get_starting_edge()->accept( this ) ;

    /**
     * Get the number of vertices of the triangulation.
     */
    nv = _vlist.size() ;

    /**
     * Allocate memory for the vertex coordinates.
     */
    verts = new double[ 2 * nv ] ;
    
    /**
     * Get the coordinates of all vertices.
     */
    VTAB::const_iterator vit ;

    for ( vit = _vlist.begin() ; vit != _vlist.end() ; ++vit ) {
      unsigned i = vit->second ;

      verts[ 2 * i     ] = vit->first->x() ;
      verts[ 2 * i + 1 ] = vit->first->y() ;
    }


    /**
     * Get the number of edges of the triangulation.
     */
    ne = _elist.size() ;

    /**
     * Allocate memory for the vertex coordinates.
     */
    edges = new unsigned[ 2 * ne ] ;
    
    /**
     * Get the index of the two vertices of each edge.
     */
    for ( unsigned i = 0 ; i < ne ; i++ ) {
      Point* p1 = _elist[ i ]->get_orig() ;
      vit = _vlist.find( p1 ) ;
      assert( vit != _vlist.end() ) ;
      edges[ 2 * i ] = vit->second ;

      Point* p2 = _elist[ i ]->get_dest() ;
      vit = _vlist.find( p2 ) ;
      assert( vit != _vlist.end() ) ;
      edges[ 2 * i + 1 ] = vit->second ;
    }

    /**
     * Unmark the mark flag of all quad-edges.
     */
    for ( unsigned i = 0 ; i < _emarked.size() ; i++ ) {
      _emarked[ i ]->get_owner()->set_mark( false ) ;
    }
    
    /**
     * Clean up the temporary vertex and edge lists.
     */
    _vlist.clear() ;
    _elist.clear() ;
    _emarked.clear() ;
    
    return ;
  }


}

/** @} */ //end of group class.
