/** 
 * \file edgevisitor.hpp
 *  
 * \brief Definition and implementation of class EdgeVisitor, which is
 * a  visitor  that  creates  a  list  of  all  edges  of  a  Delaunay
 * triangulation represented by a quad-edge data structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef EDGEVISITOR_HPP
#define EDGEVISITOR_HPP

#include "visitor.hpp"  // Visitor
#include "deltri.hpp"   // DelTri
#include "quadedge.hpp" // QuadEdge
#include "edge.hpp"     // Edge
#include "point.hpp"    // Point

#include <cmath>        // fabs
#include <map>          // map
#include <functional>   // make_pair
#include <vector>       // vector

/** 
* \defgroup QuadEdgeNameSpace Namespace quadedge.
* @{
*/

/**
* \namespace quadedge
*
* \brief   The  namespace   quadedge  contains   the   definition  and
* implementation of all classes of the quad-edge data structure, which
* can be used to  represent arbitrary planar subdivisions in Euclidian
* space.
*
*/

namespace quadedge {

  /**
   * \class EdgeVisitor
   *
   * \brief  This   class  is  a   generic  visitor  for   a  Delaunay
   * triangulation represented by a quad-edge data structure.
   *
   */
  class EdgeVisitor : public Visitor {
  public:

    // ---------------------------------------------------------------
    //
    // Type definitions
    //
    // ---------------------------------------------------------------

    /**
     * \typedef VTAB
     *
     * \brief Defines a type for a vertex identifier hash table.
     */
    typedef std::map< Point* , unsigned > VTAB ; 


    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn EdgeVisitor()
     *
     * \brief Creates an instance of this class.
     *
     */
    EdgeVisitor()
    {}


    /**
     * \fn void visit( DelTri& dt , unsigned& ne , unsigned*& edges , unsigned& nv , double*& verts )
     *
     * \brief   A  generic   visitor  method   to  visit   a  Delaunay
     * triangulation  represented  by  the quad-edge  data  structure.
     * During  the   visit,  the  method  creates  a   list  with  the
     * coordinates of  origin and destination vertices  of all visited
     * edges.
     *
     * \param dt A Delaunay triangulation represented by the quad-edge
     * data structure.
     * \param ne Total number of edges in the triangulation.
     * \param  edges An  array with  the  indices of  all edge  vertex
     * coordinates.
     * \param nv Total number of vertices of the triangulation.
     * \param verts Coordinates of the origin and destination vertices
     * of all edges.
     *
     */
    void visit( 
	       DelTri& dt ,
	       unsigned& ne ,
	       unsigned*& edges ,
	       unsigned& nv ,
	       double*& verts 
	      ) ;


    /**
     * \fn void visit ( QuadEdge* qe )
     *
     * \brief  A generic  visitor method  to  visit a  quad-edge of  a
     * Delaunay  triangulation  represented   by  the  quad-edge  data
     * structure.
     *
     * \param qe A quad-edge of the quad-edge data structure.
     *
     */
    void visit( QuadEdge* qe )
    {
      /*
       * Visit the edges of this quad-edge.
       */
      if ( !qe->is_marked() ) {
	qe->get_edge( 0 )->accept( this ) ;
	qe->get_edge( 1 )->accept( this ) ;
	qe->get_edge( 2 )->accept( this ) ;
	qe->get_edge( 3 )->accept( this ) ;
      }

      return ;
    }

    /**
     * \fn void visit( Edge* e )
     *
     * \brief A generic visitor method  to visit an edge of a Delaunay
     * triangulation represented by the quad-edge data structure.
     *
     * \param e An edge of the quad-edge data structure.
     *
     */
    virtual void visit( Edge* e )
    {
      /**
       * If  the  quad-edge  containing  this edge  has  already  been
       * visited,  then we  do  nothing.  Otherwise,  we  pick up  the
       * coordinates  of the  origin and  destination vertices  of the
       * edge, and visit edges in the adjacent faces.
       */
      if ( !e->get_owner()->is_marked() ) {
	/**
	 * Visit the two vertices incident to this edge, unless one of
	 * them is a vertex of the enclosing triangle (we ignore those
	 * vertices).
	 */
	// double mox = fabs( e->get_orig()->x() ) ;
	// double moy = fabs( e->get_orig()->y() ) ;

	e->get_orig()->accept( this ) ;
	e->get_dest()->accept( this ) ;

	//
	// Store the edge in the edge list.
	//
	_elist.push_back( e ) ;

	/*
	if ( ( mox != _max ) && ( moy != _max ) ) {
	  double mdx = fabs( e->get_dest()->x() ) ;
	  double mdy = fabs( e->get_dest()->y() ) ;

	  if ( ( mdx != _max ) && ( mdy != _max ) ) {
	    e->get_orig()->accept( this ) ;
	    e->get_dest()->accept( this ) ;

	    //
	    // Store the edge in the edge list.
	    //
	    _elist.push_back( e ) ;
	  }
	}
	*/

	/**
	 * Mark the quad-edge containing this edge as visited.
	 */
	e->get_owner()->set_mark( true ) ;

	/**
	 * Store the edge in the edge list.
	 */
	_emarked.push_back( e ) ;

	/**
	 * Visit edges in adjacent faces.
	 */
	e->o_next()->accept( this ) ;
	e->o_prev()->accept( this ) ;
	e->d_next()->accept( this ) ;
	e->d_prev()->accept( this ) ;
      }

      return ;
    }


    /**
     * \fn void visit( Point* p )
     *
     * \brief A generic visitor method to visit a vertex of a Delaunay
     * triangulation represented by the quad-edge data structure.
     *
     * \param p A vertex of the quad-edge data structure.
     *
     */
    void visit( Point* p )
    {
      VTAB::iterator vit = _vlist.find( p ) ;

      if ( vit == _vlist.end() ) {
	_vlist.insert( std::make_pair( p , _vlist.size() ) ) ;
      }

      return ;
    }


  private:

    // ---------------------------------------------------------------
    //
    // Private data members
    //
    // ---------------------------------------------------------------

    VTAB _vlist ;  ///< A temporary list of vertex coordinates.

    std::vector< Edge* > _elist ;  ///< A temporary list of edge vertex identifiers.

    std::vector< Edge* > _emarked ;  ///< A temporary list of marked edges.

    double _max ; ///< Largest absolute value of a triangulation vertex coordinate.

  } ;


}

/** @} */ //end of group class.

#endif   // EDGEVISITOR_HPP
