/** 
 * \file point.hpp
 *  
 * \brief  Definition and  implementation  of the  class Point,  which
 * represents a vertex of an edge in the quad-edge data structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

// -------------------------------------------------------------------
//
// For more  details regarding the  quad-edge data structure,  see the
// paper
//
// "Primitives for  the manipulation  of general subdivisions  and the
// computation  of Voronoi  Diagrams",  by Leonidas  Guibas and  Jorge
// Stolfi,  published  in  the  ACM Transactions  on  Graphics,  4(2),
// p. 74-123, 1985.
//
// Jorge Stolfi  made available  his own code  for the  quad-edge data
// structure:
//
// http://www.ic.unicamp.br/~stolfi/EXPORT/software/c/Index.html
//
// -------------------------------------------------------------------

#ifndef POINT_HPP
#define POINT_HPP

#include "visitor.hpp"    // Visitor

#include <list>           // std::list


/** 
* \defgroup QuadEdgeNameSpace Namespace quadedge.
* @{
*/

/**
* \namespace quadedge
*
* \brief   The  namespace   quadedge  contains   the   definition  and
* implementation of all classes of the quad-edge data structure, which
* can be used to  represent arbitrary planar subdivisions in Euclidian
* space.
*
*/

namespace quadedge {

  /**
   * Forward definition of class Edge.
   */


  /**
   * \class Point
   *
   * \brief This class represents a vertex of an edge in the quad-edge
   * data structure.  In principle,  the vertex could be in \f$E^n\f$,
   * but  our  implementation  of  the  quad-edge  data  structure  is
   * restricted to subdivisions in  \f$E^2\f$ (the 2d Euclidean affine
   * space).
   */
  class Point {
  public: 

    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn Point()
     *
     * \brief Creates an instance of this class.
     *
     */
    Point() :_x( 0 ) , _y( 0 ) , _edge( 0 ) // , _pos( 0 )
    {}


    /**
     * \fn Point( double x , double y , Edge* e )
     *
     * \brief Creates an instance of this class.
     *
     * \param x The first Cartesian coordinate of this vertex.
     * \param y The second Cartesian coordinate of this vertex.
     * \param e An edge incident to this vertex.
     *
     */
    Point( double x , double y , Edge* e ) : _x( x ) , _y( y ) , _edge( e ) // , _pos( 0 )
    {}


    /**
     * \fn Point( const Point& p )
     *
     * \brief Creates an instance  of this class from another instance
     * of this class.
     *
     * \param p An instance of this class.
     *
     */
    Point( const Point& p ) : _x( p._x ) , _y( p._y ) , _edge( p._edge ) , _pos( p._pos )
    {}


    /**
     * \fn ~Point()
     *
     * \brief Releases the memory held by an instance of this class.
     */
    virtual ~Point()
    {}


    /**
     * \fn double x() const
     *
     * \brief Returns the first Cartesian coordinate of this vertex.
     *
     * \return The first Cartesian coordinate of this vertex.
     *
     */
   double x() const
    {
      return _x ;
    }


    /**
     * \fn double y() const
     *
     * \brief Returns the second Cartesian coordinate of this vertex.
     *
     * \return The second Cartesian coordinate of this vertex.
     *
     */
   double y() const
    {
      return _y ;
    }


   /**
    * \fn Edge* get_edge() const
    *
    * \brief Returns an edge incident to this vertex.
    *
    * \return An edge incident to this vertex.
    *
    */
   Edge* get_edge() const
    {
      return _edge ;
    }


   /**
    * \fn std::list< Point* >::iterator get_position() const
    *
    * \brief  Returns  the position  of  this  point  in the  list  of
    * vertices of the triangulation it belongs to.
    *
    * \return The  position of this point  in the list  of vertices of
    * the triangulation it belongs to.
    *
    */
   std::list< Point* >::iterator get_position() const
    {
      return _pos ;
    }


    /**
     * \fn void set_coords( double x , double y )
     *
     * \brief Assigns values to the coordinates of this vertex.
     *
     * \param x The first Cartesian coordinate of this vertex.
     * \param y The second Cartesian coordinate of this vertex.
     *
     */
    void set_coords( double x , double y )
    {
      _x = x ;
      _y = y ;
    }


    /**
     * \fn void set_edge( Edge* edge )
     *
     * \brief Assigns an edge to this vertex.
     *
     * \param edge A pointer to an edge.
     *
     */
    void set_edge( Edge* edge )
    {
      _edge = edge ;
    }


    /**
     * \fn void set_position( const std::list< Point* >::iterator& it )
     *
     * \brief Assigns a position for this point in a list of points of
     * the triangulation it belongs to.
     *
     * \param  it An  iterator position  in a  list of  points  of the
     * triangulation it belongs to.
     *
     */
    void set_position( const std::list< Point* >::iterator& it )
    {
      _pos = it ;
    }


    /**
     * \fn Point& operator=( const Point& p )
     *
     * \brief Makes this vertex the same as a given vertex.
     *
     * \param p A given vertex.
     *
     */
    Point& operator=( const Point& p )
    {
      _x = p.x() ;
      _y = p.y() ;

      return *this ;
    }


    /**
     * \fn bool operator==( const Point& p ) const
     *
     * \brief Returns  the Boolean  value true if,  and only  if, this
     * vertex is the same as a given one.
     *
     * \param p A given vertex.
     *
     * \return The Boolean value true if  this vertex is the same as a
     * given one, and the Boolean value false otherwise.
     *
     */
    bool operator==( const Point& p ) const
    {
      return ( x() == p.x() ) && ( y() == p.y() ) ;
    }


    /**
     * \fn bool operator!=( const Point& p ) const
     *
     * \brief Returns  the Boolean  value true if,  and only  if, this
     * vertex is not the same as a given one.
     *
     * \param p A given vertex.
     *
     * \return The Boolean  value true if this vertex  is not the same
     * as a given one, and the Boolean value false otherwise.
     *
     */
    bool operator!=( const Point& p ) const
    {
      return !( *this == p ) ;
    }


    /**
     * \fn void accept( Visitor* v )
     *
     * \brief Accepts a visit of a generic visitor.
     *
     * \param v A pointer to a generic visitor.
     *
     */
    void accept( Visitor* v )
    {
      v->visit( this ) ;

      return ;
    }

 
  protected :

    // ---------------------------------------------------------------
    //
    // Protected data members
    //
    // ---------------------------------------------------------------

    double _x ;  ///< The first Cartesian coordinate of the vertex.
    double _y ;  ///< The second Cartesian coordinate of the vertex.
    Edge* _edge ;  ///< Pointer to an edge that has his point as its origin.
    std::list< Point* >::iterator _pos ;  ///< Position of this point in the list of vertices of the triangulation it belongs to. 
    
  } ;


}

/** @} */ //end of group class.

#endif   // POINT_HPP
