/** 
 * \file predicate.cpp
 *  
 * \brief Implementation of all static methods of the class Predicate,
 * which  represents   a  set  of  robust   predicates  for  geometric
 * computations in \f$E^2\f$.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "predicate.hpp"

#include <gmp.h>     // for all GMP functions
#include <gmpxx.h>   // for all GMP++ functions

/**
 * \fn bool Predicate::left( double px , double py , double qx , double qy , double rx , double ry )
 *
 * \brief Returns  the Boolean value  true if three given  points in
 * \f$E^2\f$   define  a  positively   oriented  basis,   and  false
 * otherwise.
 *
 * \param px First Cartesian coordinate of the first point.
 * \param py Second Cartesian coordinate of the first point.
 * \param qx First Cartesian coordinate of the second point.
 * \param qy Second Cartesian coordinate of the second point.
 * \param rx First Cartesian coordinate of the third point.
 * \param ry Second Cartesian coordinate of the third point.
 *
 * \return The Boolean value true if three given points in \f$E^2\f$
 * define a  positively oriented basis, and the  Boolean value false
 * otherwise.
 */
bool
Predicate::left(
		double px ,
		double py ,
		double qx ,
		double qy ,
		double rx ,
		double ry 
	       )
{
  return cmp( o2d( px , py , qx , qy , rx , ry ) , 0 ) > 0 ;
}


/**
 * \fn bool Predicate::left_on( double px , double py , double qx , double qy , double rx , double ry )
 *
 * \brief  Returns the  Boolean value  true if  three given  points in
 * \f$E^2\f$ are collinear or  define a positively oriented basis, and
 * false otherwise.
 *
 * \param px First Cartesian coordinate of the first point.
 * \param py Second Cartesian coordinate of the first point.
 * \param qx First Cartesian coordinate of the second point.
 * \param qy Second Cartesian coordinate of the second point.
 * \param rx First Cartesian coordinate of the third point.
 * \param ry Second Cartesian coordinate of the third point.
 *
 * \return The Boolean  value true if three given  points in \f$E^2\f$
 * are  collinear  or define  a  positively  oriented  basis, and  the
 * Boolean value false otherwise.
 */
bool
Predicate::left_on(
		   double px , 
		   double py , 
		   double qx , 
		   double qy , 
		   double rx , 
		   double ry 
		  )
{
  return cmp( o2d( px , py , qx , qy , rx , ry ) , 0 ) >= 0 ;
}


/**
 * \fn bool Predicate::collinear( double px , double py , double qx , double qy , double rx , double ry )
 *
 * \brief  Returns the  Boolean value  true if  three given  points in
 * \f$E^2\f$ are collinear, and the Boolean value false otherwise.
 *
 * \param px First Cartesian coordinate of the first point.
 * \param py Second Cartesian coordinate of the first point.
 * \param qx First Cartesian coordinate of the second point.
 * \param qy Second Cartesian coordinate of the second point.
 * \param rx First Cartesian coordinate of the third point.
 * \param ry Second Cartesian coordinate of the third point.
 *
 * \return The Boolean  value true if three given  points in \f$E^2\f$
 * are collinear, and the Boolean value false otherwise.
 */
bool
Predicate::collinear(
		     double px ,
		     double py ,
		     double qx ,
		     double qy ,
		     double rx ,
		     double ry 
		    )
{
  return cmp( o2d( px , py , qx , qy , rx , ry ) , 0 ) == 0 ;
}


/**
 * \fn Predicate::ORIENTATION Predicate::classify( double px , double py , double qx , double qy , double rx , double ry )
 *
 * \brief Returns the orientation of  a point with respect to a line
 * segment.
 *
 * \param px First Cartesian coordinate of the origin point of the line segment.
 * \param py Second Cartesian coordinate of the origin point of the line segment.
 * \param qx First Cartesian coordinate of the destination point of the line segment.
 * \param qy Second Cartesian coordinate of the destination point of the line segment.
 * \param rx First Cartesian coordinate of the point to be tested.
 * \param ry Second Cartesian coordinate of the point to be tested.
 *
 * \return The orientation of a point with respect to a line segment.
 *
 */
Predicate::ORIENTATION
Predicate::classify(
		    double px ,
		    double py ,
		    double qx ,
		    double qy ,
		    double rx ,
		    double ry 
		   )
{
  mpf_class res( 0 , 512 ) ;

  res = o2d( px , py , qx , qy , rx , ry ) ;

  if ( cmp( res , 0 ) > 0 ) {
    return LEFT ;
  }
  else if ( cmp( res , 0 ) < 0 ) {
    return RIGHT ;
  }

  if ( ( rx == px ) && ( ry == py ) ) {
    return ORIGIN ;
  }

  if ( ( rx == qx ) && ( ry == qy ) ) {
    return DESTINATION ;
  }

  mpf_class a( 0 , 64 ) ;
  mpf_class b( 0 , 64 ) ;
  mpf_class c( 0 , 64 ) ;
  mpf_class d1( 0, 512 ) ;
  mpf_class d2( 0, 512 ) ;

  if ( px != qx ) {
    a = px ; 
    b = qx ;
    c = rx ;

    d1 = c - a ;
    d2 = b - c ;

    if ( ( d1 * d2 ) > 0 ) {
      return BETWEEN ;
    }
    else if ( abs( d1 ) < abs( d2 ) ) {
      return BEHIND ;
    }

    return BEYOND ;
  }

  a = py ; 
  b = qy ;
  c = ry ;

  d1 = c - a ;
  d2 = b - c ;
  
  if ( ( d1 * d2 ) > 0 ) {
    return BETWEEN ;
  }
  else if ( abs( d1 ) < abs( d2 ) ) {
    return BEHIND ;
  }
  
  return BEYOND ;
}


/**
 * \fn bool Predicate::incircle( double px , double py , double qx , double qy , double rx , double ry , double tx , double ty )
 *
 * \brief Returns  the Boolean value true  if a given point  is in the
 * interior of  the circle  defined by three  other given  points, and
 * false otherwise.
 *
 * \param px First Cartesian coordinate of the first point defining the circle.
 * \param py Second Cartesian coordinate of the first point defining the circle.
 * \param qx First Cartesian coordinate of the second point defining the circle.
 * \param qy Second Cartesian coordinate of the second point defining the circle.
 * \param rx First Cartesian coordinate of the third point defining the circle.
 * \param ry Second Cartesian coordinate of the third point defining the circle.
 * \param tx First Cartesian coordinate of the point to be tested.
 * \param ty Second Cartesian coordinate of the point to be tested.
 *
 * \return  The  Boolean value  true  if a  given  point  is in  the
 * interior of the  circle defined by three other  given points, and
 * false otherwise.
 */
bool 
Predicate::incircle( 
		    double px ,
		    double py ,
		    double qx ,
		    double qy ,
		    double rx ,
		    double ry ,
		    double tx ,
		    double ty 
		   )
{
  mpf_class ax( px , 64 ) ;
  mpf_class ay( py , 64 ) ;
  mpf_class bx( qx , 64 ) ;
  mpf_class by( qy , 64 ) ;
  mpf_class cx( rx , 64 ) ;
  mpf_class cy( ry , 64 ) ;
  mpf_class dx( tx , 64 ) ;
  mpf_class dy( ty , 64 ) ;
  
  mpf_class res( 0 , 512 ) ;
  
  res = ( ax * ax + ay * ay ) * o2d( qx , qy , rx , ry , tx , ty )
      - ( bx * bx + by * by ) * o2d( px , py , rx , ry , tx , ty )
      + ( cx * cx + cy * cy ) * o2d( px , py , qx , qy , tx , ty )
      - ( dx * dx + dy * dy ) * o2d( px , py , qx , qy , rx , ry ) ;
  
  return cmp( res , 0 ) > 0 ;
}


/**
 * \fn mpf_class Predicate::o2d( double px , double py , double qx , double qy , double rx , double ry )
 *
 * \brief Returns twice the signed area of a triangle defined by three
 * given vertices, and the value zero if the vertices are collinear.
 *
 * \param px First Cartesian coordinate of the first point.
 * \param py Second Cartesian coordinate of the first point.
 * \param qx First Cartesian coordinate of the second point.
 * \param qy Second Cartesian coordinate of the second point.
 * \param rx First Cartesian coordinate of the third point.
 * \param ry Second Cartesian coordinate of the third point.
 *
 * \return  Twice the  signed area  of a  triangle defined  by three
 * given vertices, and the value zero if the vertices are collinear.
 */
mpf_class
Predicate::o2d(
	       double px ,
	       double py ,
	       double qx ,
	       double qy ,
	       double rx ,
	       double ry 
	      )
{
  mpf_class ax( px , 64 ) ;
  mpf_class ay( py , 64 ) ;
  mpf_class bx( qx , 64 ) ;
  mpf_class by( qy , 64 ) ;
  mpf_class cx( rx , 64 ) ;
  mpf_class cy( ry , 64 ) ;
  
  mpf_class res( 0 , 512 ) ;

  res = ( ( bx - ax ) * ( cy - ay ) ) - ( ( by - ay ) * ( cx - ax ) ) ;
  
  return res ;
}
