/** 
 * \file predicate.hpp
 *  
 * \brief Definition and implementation  of the class Predicate, which
 * represents a set of robust geometric predicates.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef PREDICATE_HPP
#define PREDICATE_HPP

#include <gmp.h>     // for all GMP functions
#include <gmpxx.h>   // for all GMP++ functions


/**
 * \class Predicate
 *
 * \brief This class represents a set of geometric predicates.
 */
class Predicate {
public:

  // -----------------------------------------------------------------
  //
  // Type definitions
  //
  // -----------------------------------------------------------------

  /**
   * \typedef ORIENTATION
   *
   * \brief A type for all possible orientation results.
   */
  typedef enum { ORIGIN, DESTINATION, LEFT, RIGHT, BEYOND, BEHIND, BETWEEN } ORIENTATION ; 


  // -----------------------------------------------------------------
  //
  // Public (static) methods
  //
  // -----------------------------------------------------------------
  
  /**
   * \fn static bool left( double px , double py , double qx , double qy , double rx , double ry )
   *
   * \brief Returns  the Boolean value  true if three given  points in
   * \f$E^2\f$   define  a  positively   oriented  basis,   and  false
   * otherwise.
   *
   * \param px First Cartesian coordinate of the first point.
   * \param py Second Cartesian coordinate of the first point.
   * \param qx First Cartesian coordinate of the second point.
   * \param qy Second Cartesian coordinate of the second point.
   * \param rx First Cartesian coordinate of the third point.
   * \param ry Second Cartesian coordinate of the third point.
   *
   * \return The Boolean value true if three given points in \f$E^2\f$
   * define a  positively oriented basis, and the  Boolean value false
   * otherwise.
   */
  static bool left(  double px , double py , double qx , double qy , double rx , double ry ) ;


  /**
   * \fn static bool left_on( double px , double py , double qx , double qy , double rx , double ry )
   *
   * \brief Returns  the Boolean value  true if three given  points in
   * \f$E^2\f$ are  collinear or  define a positively  oriented basis,
   * and false otherwise.
   *
   * \param px First Cartesian coordinate of the first point.
   * \param py Second Cartesian coordinate of the first point.
   * \param qx First Cartesian coordinate of the second point.
   * \param qy Second Cartesian coordinate of the second point.
   * \param rx First Cartesian coordinate of the third point.
   * \param ry Second Cartesian coordinate of the third point.
   *
   * \return The Boolean value true if three given points in \f$E^2\f$
   * are  collinear or  define a  positively oriented  basis,  and the
   * Boolean value false otherwise.
   */
  static bool left_on(  double px , double py , double qx , double qy , double rx , double ry ) ;


  /**
   * \fn static bool collinear( double px , double py , double qx , double qy , double rx , double ry )
   *
   * \brief Returns  the Boolean value  true if three given  points in
   * \f$E^2\f$ are collinear, and the Boolean value false otherwise.
   *
   * \param px First Cartesian coordinate of the first point.
   * \param py Second Cartesian coordinate of the first point.
   * \param qx First Cartesian coordinate of the second point.
   * \param qy Second Cartesian coordinate of the second point.
   * \param rx First Cartesian coordinate of the third point.
   * \param ry Second Cartesian coordinate of the third point.
   *
   * \return The Boolean value true if three given points in \f$E^2\f$
   * are collinear, and the Boolean value false otherwise.
   */
  static bool collinear(  double px , double py , double qx , double qy , double rx , double ry ) ;


  /**
   * \fn static ORIENTATION classify( double px , double py , double qx , double qy , double rx , double ry )
   *
   * \brief Returns the orientation of  a point with respect to a line
   * segment.
   *
   * \param px First Cartesian coordinate of the origin point of the line segment.
   * \param py Second Cartesian coordinate of the origin point of the line segment.
   * \param qx First Cartesian coordinate of the destination point of the line segment.
   * \param qy Second Cartesian coordinate of the destination point of the line segment.
   * \param rx First Cartesian coordinate of the point to be tested.
   * \param ry Second Cartesian coordinate of the point to be tested.
   *
   * \return The orientation of a point with respect to a line segment.
   *
   */
  static ORIENTATION classify( double px , double py , double qx ,
    double qy , double rx , double ry ) ;


  /**
   * \fn static bool incircle( double px , double py , double qx , double qy , double rx , double ry , double tx , double ty )
   *
   * \brief Returns the Boolean value true  if a given point is in the
   * interior of the  circle defined by three other  given points, and
   * false otherwise.
   *
   * \param px First Cartesian coordinate of the first point defining the circle.
   * \param py Second Cartesian coordinate of the first point defining the circle.
   * \param qx First Cartesian coordinate of the second point defining the circle.
   * \param qy Second Cartesian coordinate of the second point defining the circle.
   * \param rx First Cartesian coordinate of the third point defining the circle.
   * \param ry Second Cartesian coordinate of the third point defining the circle.
   * \param tx First Cartesian coordinate of the point to be tested.
   * \param ty Second Cartesian coordinate of the point to be tested.
   *
   * \return  The  Boolean value  true  if a  given  point  is in  the
   * interior of the  circle defined by three other  given points, and
   * false otherwise.
   */
  static bool incircle( double px , double py , double qx , double qy , double rx , double ry , double tx , double ty ) ;


  /**
   * \fn ~Predicate()
   *
   * \brief Releases the memory held by an instance of this class.
   */
  ~Predicate()
  {}


private :

  /**
   * \fn Predicate()
   *
   * \brief Creates an instance of this class.
   */
  Predicate() {}


  /**
   * \fn static mpf_class o2d( double px , double py , double qx , double qy , double rx , double ry )
   *
   * \brief Returns  twice the  signed area of  a triangle  defined by
   * three  given vertices,  and the  value zero  if the  vertices are
   * collinear.
   *
   * \param px First Cartesian coordinate of the first point.
   * \param py Second Cartesian coordinate of the first point.
   * \param qx First Cartesian coordinate of the second point.
   * \param qy Second Cartesian coordinate of the second point.
   * \param rx First Cartesian coordinate of the third point.
   * \param ry Second Cartesian coordinate of the third point.
   *
   * \return  Twice the  signed area  of a  triangle defined  by three
   * given vertices, and the value zero if the vertices are collinear.
   */
  static mpf_class o2d( double px , double py , double qx , 
    double qy , double rx , double ry ) ;


} ;

#endif   // PREDICATE_HPP
