/** 
 * \file quadedge.hpp
 *  
 * \brief Definition  and implementation of the  class QuadEdge, which
 * represents  two pairs  of  symmetric edges  of  the quad-edge  data
 * structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

// -------------------------------------------------------------------
//
// For more  details regarding the  quad-edge data structure,  see the
// paper
//
// "Primitives for  the manipulation  of general subdivisions  and the
// computation  of Voronoi  Diagrams",  by Leonidas  Guibas and  Jorge
// Stolfi,  published  in  the  ACM Transactions  on  Graphics,  4(2),
// p. 74-123, 1985.
//
// Jorge Stolfi  made available  his own code  for the  quad-edge data
// structure:
//
// http://www.ic.unicamp.br/~stolfi/EXPORT/software/c/Index.html
//
// -------------------------------------------------------------------

#ifndef QUADEDGE_HPP
#define QUADEDGE_HPP

#include "visitor.hpp" // Visitor
#include "edge.hpp"    // Edge

#include <cassert>     // assert


/** 
 * \defgroup QuadEdgeNameSpace Namespace quadedge.
 * @{
 */

/**
 * \namespace quadedge
 *
 * \brief   The  namespace  quadedge   contains  the   definition  and
 * implementation  of all  classes  of the  quad-edge data  structure,
 * which  can be used  to represent  arbitrary planar  subdivisions in
 * Euclidian space.
 *
 */

namespace quadedge {

  /**
   * \class QuadEdge
   *
   * \brief This class represents two  pairs of symmetric edges of the
   * quad-edge data structure.
   */
  class QuadEdge {
  public:

    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn QuadEdge()
     *
     * \brief Creates an instance of this class.
     *
     */
    QuadEdge()
    {
      /**
       * Set up the id, quad-edge, and origin vertex of each edge.
       */
      for ( unsigned i = 0 ; i <= 3 ; i++ ) {
	_edges[ i ].set_owner( this ) ;
	_edges[ i ].set_id( i ) ;
      }

      /**
       * Initialize the  "next" field of  the four edges.  Recall that
       * the two symmetric edges of the dual are "next" of each other,
       * while  the  symmetric  edges  of  the primal  are  "next"  of
       * themselves.
       */
      _edges[ 0 ].set_next( &( _edges[ 0 ] ) ) ;   // primal edge
      _edges[ 2 ].set_next( &( _edges[ 2 ] ) ) ;   // prinal edge

      _edges[ 1 ].set_next( &( _edges[ 3 ] ) ) ;   // dual edge
      _edges[ 3 ].set_next( &( _edges[ 1 ] ) ) ;   // dual edge

      /**
       * Set the mark flag as false.
       */
      set_mark( false ) ;

      return ;
    }


    /**
     * \fn ~QuadEdge()
     *
     * \brief Releases the memory held by an instance of this class.
     */
    virtual ~QuadEdge()
    {}


    /**
     * \fn Edge* get_edge( unsigned i )
     *
     * \brief Returns the i-th edge of this quad-edge.
     *
     * \param i An edge index.
     *
     * \return The i-th edge of this quad-edge.
     *
     */
    Edge* get_edge( unsigned i )
    {
      assert( i <= 3 ) ;

      return &( _edges[ i ] ) ;
    }


    /**
     * \fn void set_mark( bool mark )
     *
     * \brief  Assigns  a Boolean  value  to  the  mark flag  of  this
     * quad-edge.
     *
     * \param mark A Boolean value.
     *
     */
    void set_mark( bool mark )
    {
      _marked = mark ;
    }


    /**
     * \fn bool is_marked() const
     *
     * \brief  Returns the Boolean  value true  if this  quad-edge has
     * been marked, and the Boolean value false otherwise.
     *
     * \return  The Boolean  value  true if  this  quad-edge has  been
     * marked, and the Boolean value false otherwise.
     *
     */
    bool is_marked() const
    {
      return _marked ;
    }


    /**
     * \fn void accept( Visitor* v )
     *
     * \brief Accepts a visit of a generic visitor.
     *
     * \param v A pointer to a generic visitor.
     *
     */
    void accept( Visitor* v )
    {
      v->visit( this ) ;

      return ;
    }


  protected:

    // ---------------------------------------------------------------
    //
    // Protected data members
    //
    // ---------------------------------------------------------------

    Edge _edges[ 4 ] ;  ///< Array with the two pairs of symmetric edges that define this quad-edge.

    bool _marked ;      ///< A flag to be used by visitor-type of algorithms.

  } ;


}

/** @} */ //end of group class.

#endif   // QUADEDGE_HPP
