/** 
 * \file visitor.hpp
 *  
 * \brief Definition of the abstract class Visitor, which is a generic
 * visitor  for a  Delaunay triangulation  represented by  a quad-edge
 * data structure.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef VISITOR_HPP
#define VISITOR_HPP


/** 
* \defgroup QuadEdgeNameSpace Namespace quadedge.
* @{
*/

/**
* \namespace quadedge
*
* \brief   The  namespace   quadedge  contains   the   definition  and
* implementation of all classes of the quad-edge data structure, which
* can be used to  represent arbitrary planar subdivisions in Euclidian
* space.
*
*/

namespace quadedge {

  /**
   * Forward declaration of classes.
   */
  class QuadEdge ;
  class Edge ;
  class Point ;


  /**
   * \class Visitor
   *
   * \brief  This   class  is  a   generic  visitor  for   a  Delaunay
   * triangulation represented by a quad-edge data structure.
   *
   */
  class Visitor {
  public:

    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn Visitor()
     *
     * \brief Creates an instance of this class.
     *
     */
    Visitor()
    {}


    /**
     * \fn virtual void visit ( QuadEdge* qe ) = 0
     *
     * \brief  A generic  visitor method  to  visit a  quad-edge of  a
     * Delaunay  triangulation  represented   by  the  quad-edge  data
     * structure.
     *
     * \param  qe A  pointer  to  a quad-edge  of  the quad-edge  data
     * structure.
     *
     */
    virtual void visit ( QuadEdge* qe ) = 0 ;


    /**
     * \fn virtual void visit ( Edge* e ) = 0
     *
     * \brief A generic visitor method  to visit an edge of a Delaunay
     * triangulation represented by the quad-edge data structure.
     *
     * \param e A pointer to an edge of the quad-edge data structure.
     *
     */
     virtual void visit ( Edge* e ) = 0 ;


    /**
     * \fn virtual void visit ( Point* p ) = 0
     *
     * \brief A generic visitor method to visit a vertex of a Delaunay
     * triangulation represented by the quad-edge data structure.
     *
     * \param p A pointer to a vertex of the quad-edge data structure.
     *
     */
     virtual void visit ( Point* p ) = 0 ;


  } ;


}

/** @} */ //end of group class.

#endif   // VISITOR_HPP
