/** 
 * \file test-deltri.cpp
 *  
 * \brief A simple driver to test the DELTRI library.
 *
 * \author
 * Marcelo Ferreira Siqueira
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * mfsiqueira at gmail (dot) com
 *
 * \version 1.0
 * \date May 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include <iostream>                 // std::cout, std::endl
#include <string>                   // std::string
#include <cstdlib>                  // exit, atoi, EXIT_SUCCESS, EXIT_FAILURE
#include <cmath>                    // std::pow
#include <iomanip>                  // std::setprecision

#include "deltri.hpp"               // quadedge::DelTri
#include "edgevisitor.hpp"          // quadedge::EdgeVisitor

using namespace std ;

/**
 * \fn int main( int argc , char* argv[] )
 *
 * \brief Computes the Delaunay triangulation of given point set.
 *
 * \param argc The number of command line arguments.
 * \param argv The command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {
  using std::cin  ;
  using std::cout ;
  using std::endl ;

  using quadedge::DelTri ;
  using quadedge::EdgeVisitor ;

  /*
   * Read in the size of the input point set. 
   */

  if ( argc != 2 ) {
    cout << "Usage: " 
         << endl
         << "\t\t test-deltri arg1 < arg2 [ > arg3 ]" 
	 << endl
         << "\t\t arg1: the size of the point set"
         << endl 
         << "\t\t arg2: name of the input file"
         << endl
         << "\t\t arg3: name of the output file"
         << endl
         << endl ;

    return EXIT_FAILURE ;
  }

  unsigned num_pts = (unsigned) atoi( argv[ 1 ] ) ;

  if ( num_pts < 3 ) {
    cout << "Point set size must be at least 3" << endl ;
    return EXIT_FAILURE ;
  }

  /*
   * Read in the Cartesian coordinates of the points.
   */

  double* pts = new double[ 2 * num_pts ] ;

  for ( unsigned i = 0 ; i < num_pts ; i++ ) {
    cin >> pts[ 2 * i     ] ;
    cin >> pts[ 2 * i + 1 ] ;
  }

  /**
   * Compute the Delaunay triangulation of the given points.
   */

  DelTri* my_dt = new DelTri( num_pts , &pts[ 0 ] ) ;

  /**
   * Obtain all edges and vertices of the triangulation.
   */

  unsigned nv , ne ;
  unsigned* edges = 0 ;
  double* verts = 0 ;

  EdgeVisitor visitor ;
  visitor.visit( *my_dt , ne , edges , nv , verts ) ;

  /**
   * Printout the edge and vertex information.
   */

  cout << std::setprecision( 18 ) << std::fixed ;

  cout << ne << '\t' << endl ;

  for ( unsigned i = 0 ; i < ne ; i++ ) {
    cout << edges[ 2 * i     ] 
	 << '\t' 
         << edges[ 2 * i + 1 ]
         << endl ;
  }

  cout << nv << '\t' << endl ;

  for ( unsigned i = 0 ; i < nv ; i++ ) {
    cout << verts[ 2 * i     ] 
	 << '\t' 
         << verts[ 2 * i + 1 ]
         << endl ;
  }

  if ( edges != 0 ) delete[] edges ;
  if ( verts != 0 ) delete[] verts ;

  delete my_dt ;

  return EXIT_SUCCESS ;
}
